import argparse
import sys
import struct
import time
import socket
import fcntl
import termios
import signal

from icm20649_spi.spitool import SPIConn
from icm20649_spi.icm20649 import ICM20649

from . import linesocket
from . import log
from . import queue

from icm20649_spi.timer import WTimer

maxReadBlkSize = 1024
blksz = 12

class Status:
  done = False
  fcnt = wmfl = ovfl = 0
  maxfcnt = avgfcnt = stdfcnt = nfifodata = 0
  nsamples = nreads = nresets = nwmfl = 0
  curTempVals = [0]

status = Status()
timer = WTimer()

def getFIFO(icm20649):
  fcnt, wmfl, ovfl = icm20649.FIFOState
  rcnt = int(min(fcnt,maxReadBlkSize)/blksz)*blksz
  if rcnt > 0:
    v = icm20649.getFIFOBytes(rcnt)
    fifodata = struct.pack('B'*rcnt, *v)
    status.fcnt = fcnt
    status.maxfcnt = max(fcnt, status.maxfcnt)
    # assert(len(fifodata) == rcnt)
    # assert(len(v) == rcnt)
    # assert(rcnt % blksz == 0)
  else:
    fifodata = b''

  if ovfl:
    icm20649.resetFIFO = 1
    icm20649.resetFIFO = 0
    timer.click()
    print("FIFO overflow: %s, FIFO %d, buf %d" % (timer, fcnt, len(fifodata)))
    status.nresets += 1
    status.maxfcnt = 0
    sys.stdout.flush()
  elif wmfl:
    status.nwmfl += 1
#      log.error("FIFO watermark reached: it %d/%d, FIFO %d, buf %d" % (iloop, fetch, fcnt, len(fifodata)))

  return fifodata


def readFIFO(sensor, ls):
  global blksz
  timer.reset()

  blksz = sensor.getFIFOBlockSize()

  log.msg('blksz', blksz)
  
#  print(dir(socket))
#  print(dir(termios))
  sensor.resetFIFO = 0

  nfifodata = 0

  if sensor.tempmanrfreq > 0:
    tempManInter = 1 / sensor.tempmanrfreq
    status.curTempVals = [sensor.readTemp(),]

  que = queue.Queue(1e6)

  while not status.done:
    fifodata = getFIFO(sensor)
    nfifodata += len(fifodata)

    timer.click()

    if len(fifodata) > 0:
      try:
        que.put(fifodata)
      except:
        pass
    elif que.empty():
      time.sleep(4e-4)

    for i in range(min(3,que.qsize())):
      try:
        item = que.peek()
        ls.transmitb(item)
        que.get()
      except BlockingIOError as ex:
        #print('socket blocked for queue')
        break
      except BrokenPipeError as ex:
        print('socket closed, exit')
        status.done = True
        break

    if status.fcnt < 1000 and sensor.tempmanrfreq > 0 and timer.elapsed('tempfetch', tempManInter):
      status.curTempVals = [sensor.readTemp(),]

    if timer.elapsed('info', 1) or timer.last > 20:
      nwrite = 0
      # nwrite = ls.sock.getsockopt(socket.SOL_SOCKET, socket.SO_NWRITE, 8)
      print("%s%s: %d B, %.2f KB/s, FIFO %d (max %d), ov %d wm %d, queue %d" %
            (">" if timer.last > 40 else " ", timer,
             nfifodata, 1e-3 * nfifodata / timer.elaps,
             status.fcnt, status.maxfcnt, status.nresets, status.nwmfl, que.qsize()))
      sys.stdout.flush()

def ctrlCHandler(sig, frame):
  print('Signal %d caught, setting stop' % (sig,))
  status.done = True
  return

def run():
  import os
  log.name = 'fiforeader'

  if os.geteuid() == 0:
    res = os.setpriority(os.PRIO_PROCESS, 0, -20)
    print('Am root', res)

  parser = argparse.ArgumentParser(description='FIFO reader helper process')
  
  parser.add_argument('--bus', dest='bus', type=int, default=0, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, default=0, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, default=int(1e6), help='SPI speed')
  
  parser.add_argument('--outfile', type=str, help='name of output file')
  
  args = parser.parse_args(sys.argv[1:])
  print(args)

  signal.signal(signal.SIGINT, ctrlCHandler)
  signal.signal(signal.SIGTERM, ctrlCHandler)

  spi = SPIConn(args.bus, args.dev, args.speed)
  print(spi)

  icm20649 = ICM20649(spi)
  print('open icm20649, bank is', icm20649.bank)

  print(icm20649)

  if not icm20649.connected:
    log.error("icm20649 device not found")
    return
  else:
    log.msg("icm20649 device connected")
    
  ls = linesocket.LineSocket(args.outfile)
  ls.openSocket(1)

  try:
    readFIFO(icm20649, ls)
  except KeyboardInterrupt as ex:
    print('FifoReader caught Ctrl-C')
    pass
  except BaseException as ex:
    raise(ex)

  print('FifoReader exiting')

if __name__ == "__main__":
    run()
