#! /usr/bin/env python3

from icm20649_spi.spitool import SPIConn
from . import log

import time
import argparse
import math
import struct
import os

regWhoami = 0
valWhoami = 0xE1

regReset = 0x06
valReset = 0x41

def bytes2words(data):
  return struct.unpack('>' + 'h'*int(len(data)/2), data)

class ICM20649:

  low_power_mode = False

  fifo_accel = True
  fifo_gyro = True
  fifo_mag = False
  fifo_temp = False

  temp_disable  = False

  accel_gyro_freq_nom     = 1125

  i2c_cycle      = False
  accel_cycle    = False
  gyro_cycle     = False

  enable_odr_align=False

  raw_values = False

  gyro_disable   = False
  gyro_scale     = 2000  # +/- 500, 1000, 2000, 4000 dps
  gyro_filter    = 2    # 0..7
  gyro_filter_bw = 154.3
  gyro_smplrt_div= 2    # 0..255
  gyro_avg       = 1   # 1, 2, 4, 8, ..., 128
  gyro_freq      = 1100
  gyro_use_filter= True
  gyro_factor    = 1

  accel_disable  = False
  normalEarthAcc = 9.18 # m/s^2
  accel_g        = True
  accel_scale    = 8 # 4, 8, 16, 30 g
  accel_filter   = 2 # 0..7
  accel_filter_bw= 500
  accel_smplrt_div=2 # 0..4095
  accel_freq     = 1100
  accel_avg      = 4 # 1 (filter off), or 4, 8, 16, 32 (filter on)
  accel_use_filter=True
  accel_factor   = 1
  accelx_factor  = 1
  accely_factor  = 1
  accelz_factor  = 1

  # internal
  temp_filter_bw = 500

  # external
  tempmanrfreq = 0
  temp_ext_enable = [0,1]

  def __init__(self, spi, args={}):
    self.spi = spi
    for name in args:
      if hasattr(self, name):
        print('Setting property: %s=%s' % (name, args[name]))
        setattr(self, name, args[name])

    if self.gyro_freq > 0:
      gdiv = round(self.accel_gyro_freq_nom / self.gyro_freq)
      if gdiv > 0:
        gdiv -= 1
      self.gyro_smplrt_div = gdiv
    else:
      self.gyro_disable = True

    if self.accel_freq > 0:
      adiv = round(self.accel_gyro_freq_nom / self.accel_freq)
      if adiv > 0:
        adiv -= 1
      self.accel_smplrt_div = adiv
    else:
      self.accel_disable = True

    if self.accel_disable:
      self.fifo_accel = False
    if self.gyro_disable:
      self.fifo_gyro = False

    self.accel_ax_factors  = [self.accelx_factor, self.accely_factor, self.accelz_factor]
    self.bank = 0

  def __str__(self):
    self.bank = 0
    time.sleep(2e-4)
    s = "ICM20649 model\n"
    s += "whoami register:       0x%02X (%s)\n" % (self.whoami(), ("C" if self.connected else "Not c") + "onnected")
    s += "clock source:          % 5s\n" % (self.clockSource, )
    s += "timebase corr.:        % 5s %% (%d)\n" % (self.timebaseCorrection, self.timebaseCorrectionRaw)
    s += "sleep mode:            % 5s\n" % (self.enableSleep, )
    s += "low power mode:        % 5s\n" % (self.enableLowPower, )
    s += "low power DMP:         % 5s\n" % (self.lowPowerDMP, )
    s += "disable accel:         % 5s (%s)\n" % (self.disableAccel, bin(self.get(self.PWR_MGNT_2)))
    s += "disable gyro:          % 5s (%s)\n" % (self.disableGyro, bin(self.get(self.PWR_MGNT_2)))
    s += "disable temperature:   % 5s\n" % (self.disableTemperature, )

    s += '\n'
    s += "enable I2C master:     %s\n" % (self.enableI2CMaster, )
    s += "enable DMP:            %s\n" % (self.enableDMP, )
    s += "enable FIFO:           %s\n" % (self.enableFIFO, )
    s += "FIFO state:            %s\n" % (self.FIFOState, )
    s += "enable Accel -> FIFO:  % 5s (=%s)\n" % (self.enableAccelFIFO,  bin(self.get(self.FIFO_EN_2)), )
    s += "enable GyroX -> FIFO:  % 5s (=%s)\n" % (self.enableGyroXFIFO,  bin(self.get(self.FIFO_EN_2)), )
    s += "enable GyroY -> FIFO:  % 5s (=%s)\n" % (self.enableGyroYFIFO,  bin(self.get(self.FIFO_EN_2)), )
    s += "enable GyroZ -> FIFO:  % 5s (=%s)\n" % (self.enableGyroZFIFO,  bin(self.get(self.FIFO_EN_2)), )
    s += "enable Temp  -> FIFO:  % 5s (=%s)\n" % (self.enableTempFIFO,   bin(self.get(self.FIFO_EN_2)), )

    s += '\n'
    s += "enable I2C cycle:      %s\n" % (self.enableI2CCycle, )
    s += "enable Accel cycle:    %s\n" % (self.enableAccelCycle, )
    s += "enable Gyro cycle:     %s\n" % (self.enableGyroCycle, )
    s += "user bank:             %s\n" % (self.bank, )

    s += '\nSample rate      =  %g Hz\n' % (self.sampleRate,)

    s += '\nGyro freq        =  %g Hz (div: %d)' % (self.gyroFreq,self.gyroSampleRateDivider+1)
    s += '\nGyro scale       =  %g dps (=0x%02x)' % (self.gyroScale,self.gyroScaleConfig)
    s += '\nGyro use filter  =  %g' % (self.gyroUseFilter,)
    s += '\nGyro filter      =  %g Hz (=0x%02x)' % (self.gyroFilterBW,self.gyroFilterConfig)
    s += '\nGyro avg         =  %g (=0x%02x)' % (self.gyroAvg,self.gyroAvgConfig)
    s += '\nGyro offs.       =  %s => %s' % (self.gyroOffsets, self.gyro_offsets)

    s += '\n'
    s += '\nAccel freq       =  %g Hz (div: %d)' % (self.accelFreq,self.accelSampleRateDivider+1)
    s += '\nAccel scale      =  %g g (=0x%02x)' % (self.accelScale,self.accelScaleConfig)
    s += '\nAccel use filter =  %g' % (self.accelUseFilter,)
    s += '\nAccel filter     =  %g Hz (=0x%02x)' % (self.accelFilterBW,self.accelFilterConfig)
    s += '\nAccel avg        =  %g (=0x%02x)' % (self.accelAvg,self.accelAvgConfig)
    s += '\nAccel offs.      =  %s => %s' % (self.accelOffsets, self.accel_offsets)
    s += '\nAccel unit.      =  %s' % ("g" if self.accel_g else "m/s^2")

    s += '\n'
    s += '\nTemp freq        =  %g Hz (=0x%02x)' % (self.tempFreq,self.tempConfig)
    s += '\nTemp filter      =  %g Hz (=0x%02x)' % (self.tempFilterBW,self.tempConfig)

    s += '\n'
    s += '\nTemp man freq    =  %g Hz' % (self.tempmanrfreq,)
    s += '\nTemp. enabled    =  %s' % (self.temp_ext_enable,)

    s += '\n'
    s += '\nI2C ODR          =  %g Hz (=0x%02x)' % (self.i2cMasterODR,self.i2cMasterODRConfig)
    s += '\nenable I2C Multi master =%d' % (self.enableI2CMultiMaster,)
    s += '\nI2C Master Trans. mode =%s' % (self.enableI2CStopRestart,)
    s += '\nI2C_MST_ODR_CONFIG=       (=0x%02x)' % (self.getb(3, self.I2C_MST_ODR_CONFIG),)
    s += '\nI2C_MST_CLK      =        (=0x%02x)' % (self.i2cMasterClockConfig,)
    s += '\nI2C_MST_CTRL     =        (=0x%02x)' % (self.getb(3, self.I2C_MST_CTRL),)
    s += '\nI2C 0 ADDR       =        (=0x%02x)' % (self.getb(3, self.I2C_SLV0_ADDR),)
    s += '\nI2C 0 REG        =        (=0x%02x)' % (self.getb(3, self.I2C_SLV0_REG),)
    s += '\nI2C 0 CTRL       =        (=0x%02x)' % (self.getb(3, self.I2C_SLV0_CTRL),)
    s += '\nODR_ALIGN_EN     = %s (=0x%02x)' % (self.enableODRAlign, self.getb(2, self.ODR_ALIGN_EN),)

    self.bank = 0
    return s

  def get(self, addr):
    return self.spi.get(addr)

  def getb(self, bank, addr):
    assert(bank != 0)
    assert(self.bank == 0)
    self.bank = bank
    val = self.spi.get(addr)
    self.bank = 0
    return val

  def get2(self, addr):
    return self.spi.get2(addr)

  def set(self, addr, data):
    return self.spi.set(addr, data)

  def setb(self, bank, addr, data):
    assert(bank != 0)
    assert(self.bank == 0)
    self.bank = bank
    res = self.spi.set(addr, data)
    self.bank = 0
    return res

  def setr(self, addr, data):
    return self.spi.setr(addr, data)

  def reset(self):
    self.bank = 0
    if not self.connected:
      log.error("icm20649.reset: device not found")
      return
    print("device reset")
    self.set(regReset, 0xC1)
    time.sleep(1e-2)
    resetval = self.get(regReset)
    if resetval != 0x41:
      log.error("icm20649.reset: device reset failed: 0x%X" % (resetval,))

  def setBit(self, reg, bit, b=True):
    val = self.get(reg)
    bv = 1 << bit
    if int(b):
      val |= bv
    else:
      val &= ~bv
    self.set(reg, val)

  def getBit(self, reg, bit):
    val = self.get(reg)
    bv = 1 << bit
    return True if (val & bv) else False

  USER_CTRL = 0x03
  LP_CONFIG = 0x05
  PWR_MGNT_1 = 0x06
  PWR_MGNT_2 = 0x07

  @property
  def clockSource(self):
    self.bank = 0
    return self.get(self.PWR_MGNT_1) & 0x7
  @clockSource.setter
  def clockSource(self, v):
    self.bank = 0
    curv = self.get(self.PWR_MGNT_1)
    curv &= ~0x7
    curv |= v
    self.set(self.PWR_MGNT_1, curv)

  TIMEBASE_CORRECTION_PLL = 0x28
  @property
  def timebaseCorrectionRaw(self):
    return self.getb(1, self.TIMEBASE_CORRECTION_PLL)

  @property
  def timebaseCorrection(self):
    return self.timebaseCorrectionRaw * 0.079

  @property
  def disableTemperature(self):
    self.bank = 0
    return self.getBit(self.PWR_MGNT_1, 3)
  @disableTemperature.setter
  def disableTemperature(self, b):
    self.bank = 0
    self.setBit(self.PWR_MGNT_1, 3, b)

  @property
  def enableLowPower(self):
    self.bank = 0
    return self.getBit(self.PWR_MGNT_1, 5)
  @enableLowPower.setter
  def enableLowPower(self, b):
    self.bank = 0
    self.setBit(self.PWR_MGNT_1, 5, b)

  @property
  def enableSleep(self, b=True):
    self.bank = 0
    return self.getBit(self.PWR_MGNT_1, 6)
  @enableSleep.setter
  def enableSleep(self, b):
    self.bank = 0
    self.setBit(self.PWR_MGNT_1, 6, b)

  @property
  def disableGyro(self):
    self.bank = 0
    val = self.get(self.PWR_MGNT_2)
    return (val & 0b111) != 0
  @disableGyro.setter
  def disableGyro(self, b=True):
    self.bank = 0
    val = self.get(self.PWR_MGNT_2)
    if b:
      val |= 0b00000111
    else:
      val &= ~0b00000111
    self.set(self.PWR_MGNT_2, val)

  @property
  def disableAccel(self):
    self.bank = 0
    val = self.get(self.PWR_MGNT_2)
    return ((val >> 3) & 0b111) != 0
  @disableAccel.setter
  def disableAccel(self, b):
    self.bank = 0
    val = self.get(self.PWR_MGNT_2)
    if b:
      val |= 0b00111000
    else:
      val &= ~0b00111000
    self.set(self.PWR_MGNT_2, val)

  @property
  def enableDMP(self):
    self.bank = 0
    return self.getBit(self.USER_CTRL, 7)
  @enableDMP.setter
  def enableDMP(self, b):
    self.bank = 0
    self.setBit(self.USER_CTRL, 7, b)

  @property
  def enableFIFO(self):
    self.bank = 0
    return self.getBit(self.USER_CTRL, 6)
  @enableFIFO.setter
  def enableFIFO(self, b):
    self.bank = 0
    self.setBit(self.USER_CTRL, 6, b)

  @property
  def enableI2CMaster(self):
    self.bank = 0
    return self.getBit(self.USER_CTRL, 5)
  @enableI2CMaster.setter
  def enableI2CMaster(self, b):
    self.bank = 0
    self.setBit(self.USER_CTRL, 5, b)

  def enableI2CReset(self, b=True):
    self.bank = 0
    self.setBit(self.USER_CTRL, 1, b)

  @property
  def enableI2CCycle(self):
    self.bank = 0
    return self.getBit(self.LP_CONFIG, 6)
  @enableI2CCycle.setter
  def enableI2CCycle(self, b):
    self.bank = 0
    self.setBit(self.LP_CONFIG, 6, b)

  @property
  def enableAccelCycle(self):
    return self.getBit(self.LP_CONFIG, 5)
  @enableAccelCycle.setter
  def enableAccelCycle(self, b):
    self.setBit(self.LP_CONFIG, 5, b)

  @property
  def enableGyroCycle(self):
    return self.getBit(self.LP_CONFIG, 4)
  @enableGyroCycle.setter
  def enableGyroCycle(self, b):
    self.setBit(self.LP_CONFIG, 4, b)

  MOD_CTRL_USR = 0x54
  @property
  def lowPowerDMP(self):
    self.bank = 2
    return self.getBit(self.MOD_CTRL_USR, 0)
  @lowPowerDMP.setter
  def lowPowerDMP(self, b):
    self.bank = 2
    return self.setBit(self.MOD_CTRL_USR, 0, b)

  def setUserBank(self, i):
    self.set(0x7F, (i & 0x3)<<4)
  def getUserBank(self):
    return (self.get(0x7F)>>4) & 0x3

  @property
  def bank(self):
    return self.getUserBank()
  @bank.setter
  def bank(self, i):
    self.setUserBank(i)

  TEMP_CONFIG    = 0x53
  @property
  def tempConfig(self):
    return self.getb(2, self.TEMP_CONFIG) & 0x7
  @tempConfig.setter
  def tempConfig(self, value):
    cur = self.getb(2, self.TEMP_CONFIG)
    cur &= ~0x7
    cur |= (value & 0x7)
    self.setb(2, self.TEMP_CONFIG, cur)

  tempFilterBWs = [7932., 217.9, 123.5, 65.9, 34.1, 17.3, 8.8, 7932.0]
  @property
  def tempFilterBW(self):
    cval = self.tempConfig
    return self.tempFilterBWs[cval]
  @tempFilterBW.setter
  def tempFilterBW(self, value):
    (best, ind) = self.findClosest(self.tempFilterBWs, value)
    self.tempConfig = ind

  @property
  def tempFreq(self):
    cval = self.tempConfig
    if cval == 0 or cval == 7:
      return 9000
    else:
      return 1125


  def getGyroScaleConfigBits(self, scale):
    if scale == 500:
      return 0b00
    if scale == 1000:
      return 0b01
    if scale == 2000:
      return 0b10
    if scale == 4000:
      return 0b11

  gyroScales = [500, 1000, 2000, 4000]

  @property
  def gyroFreq(self):
    if self.gyroUseFilter:
      return self.accel_gyro_freq_nom/(1 + self.gyroSampleRateDivider)
    elif self.enableLowPower:
      return min(526.5, self.accel_gyro_freq_nom/(1 + self.gyroSampleRateDivider))
    else:
      return 9e3

  @property
  def gyroSampleRateDivider(self):
    return self.getb(2, self.GYRO_SMPLRT_DIV)
  @gyroSampleRateDivider.setter
  def gyroSampleRateDivider(self, val):
    self.setb(2, self.GYRO_SMPLRT_DIV, val)

  @property
  def gyroUseFilter(self):
    return self.getb(2, self.GYRO_CONFIG_1) & 0x1
  @gyroUseFilter.setter
  def gyroUseFilter(self, value):
    cur = self.getb(2, self.GYRO_CONFIG_1)
    cur &= ~0x1
    cur |= value & 0x1
    self.setb(2, self.GYRO_CONFIG_1, cur)

  @property
  def gyroScaleConfig(self):
    return (self.getb(2, self.GYRO_CONFIG_1) >> 1) & 0x3
  @gyroScaleConfig.setter
  def gyroScaleConfig(self, value):
    cur = self.getb(2, self.GYRO_CONFIG_1)
    cur &= ~0x110
    cur |= value << 1
    self.setb(2, self.GYRO_CONFIG_1, cur)

  @property
  def gyroScale(self):
    sconfig = self.gyroScaleConfig
    return self.gyroScales[sconfig]
  @gyroScale.setter
  def gyroScale(self, gyro_scale):
    self.gyroScaleConfig = self.getGyroScaleConfigBits(gyro_scale)

  @property
  def gyroFilterConfig(self):
    return (self.getb(2, self.GYRO_CONFIG_1) >> 3) & 0x7
  @gyroFilterConfig.setter
  def gyroFilterConfig(self, value):
    cur = self.getb(2, self.GYRO_CONFIG_1)
    cur &= ~0x11111000
    cur |= (value & 0x7) << 3
    self.setb(2, self.GYRO_CONFIG_1, cur)

  def findClosest(self, vals, v):
    diffPairs = [(abs(w - v),i) for (i,w) in enumerate(vals)]
    minPair = min(diffPairs, key=lambda x: x[0])
    return minPair

  gyroFilterBWs = [229.8, 187.6, 154.3, 73.3, 35.9, 17.8, 8.9, 376.5]
  @property
  def gyroFilterBW(self):
    return self.gyroFilterBWs[self.gyroFilterConfig]
  @gyroFilterBW.setter
  def gyroFilterBW(self, value):
    (best, ind) = self.findClosest(self.gyroFilterBWs, value)
    self.gyroFilterConfig = ind

  @property
  def gyroAvgConfig(self):
    return self.getb(2, self.GYRO_CONFIG_2)
  @gyroAvgConfig.setter
  def gyroAvgConfig(self, value):
    return self.setb(2, self.GYRO_CONFIG_2, value)

  @property
  def gyroAvg(self):
    return 2**self.gyroAvgConfig
  @gyroAvg.setter
  def gyroAvg(self, value):
    self.gyroAvgConfig = int(math.log2(value)) & 0x7

  def setGyroConfig(self):
    self.gyroSampleRateDivider = self.gyro_smplrt_div
    print('gyroSampleRateDivider', self.gyroSampleRateDivider)
    self.gyroScale = self.gyro_scale
    print('gyroScale', self.gyroScale)
    if self.raw_values:
      self.gyro_factor = 1
    else:
      self.gyro_factor = 2 * self.gyroScale / (2**16)
    self.gyroFilterBW = self.gyro_filter_bw
    self.gyroUseFilter = self.gyro_use_filter
    self.gyroAvg = self.gyro_avg
    self.setupGyroCorrection()

  def readGyro(self):
    ax, ay, az = self.spi.get3x2(0x33)
    ax = self.getGyroValueUnCorr(ax - self.gyro_offsets[0])
    ay = self.getGyroValueUnCorr(ay - self.gyro_offsets[1])
    az = self.getGyroValueUnCorr(az - self.gyro_offsets[2])
    return (ax, ay, az)

  def getGyroValueUnCorr(self, val):
    val *= self.gyro_factor
    return val

  def getGyroValues(self, vals):
    return [(vals[i] - self.gyro_offsets[i]) * self.gyro_factor for i in range(3)]


  def getAccelScaleConfigBits(self, scale):
    if scale == 4:
      return 0
    if scale == 8:
      return 1
    if scale == 16:
      return 2
    if scale == 32 or scale == 30:
      return 3
    print("Error: accel scale: %g is not available" % (scale,))
    return 0

  accelScales = [4, 8, 16, 32]

  @property
  def sampleRate(self):
    if not self.disableGyro:
      return self.gyroFreq
    elif not self.disableAccel:
      return self.accelFreq
    else:
      return 1125

  @property
  def accelFreq(self):
    if self.accelUseFilter:
      return self.accel_gyro_freq_nom/(1 + self.accelSampleRateDivider)
    else:
      return 4.5e3

  @property
  def accelSampleRateDivider(self):
    hb = self.getb(2, self.ACCEL_SMPLRT_DIV_1) & 0xF
    lb = self.getb(2, self.ACCEL_SMPLRT_DIV_2)
    return (hb << 8) | lb
  @accelSampleRateDivider.setter
  def accelSampleRateDivider(self, val):
    hb = (val >> 8) & 0xF
    lb = val & 0xFF
    self.setb(2, self.ACCEL_SMPLRT_DIV_1, hb)
    self.setb(2, self.ACCEL_SMPLRT_DIV_2, lb)

  @property
  def accelUseFilter(self):
    return self.getb(2, self.ACCEL_CONFIG_1) & 0x1
  @accelUseFilter.setter
  def accelUseFilter(self, value):
    cur = self.getb(2, self.ACCEL_CONFIG_1)
    cur &= ~0x1
    cur |= value & 0x1
    self.setb(2, self.ACCEL_CONFIG_1, cur)

  @property
  def accelScaleConfig(self):
    return (self.getb(2, self.ACCEL_CONFIG_1) >> 1) & 0x3
  @accelScaleConfig.setter
  def accelScaleConfig(self, value):
    cur = self.getb(2, self.ACCEL_CONFIG_1)
    cur &= ~0x110
    cur |= value << 1
    self.setb(2, self.ACCEL_CONFIG_1, cur)

  @property
  def accelScale(self):
    sconfig = self.accelScaleConfig
    return self.accelScales[sconfig]
  @accelScale.setter
  def accelScale(self, accel_scale):
    self.accelScaleConfig = self.getAccelScaleConfigBits(accel_scale)

  @property
  def accelFilterConfig(self):
    return (self.getb(2, self.ACCEL_CONFIG_1) >> 3) & 0x7
  @accelFilterConfig.setter
  def accelFilterConfig(self, value):
    cur = self.getb(2, self.ACCEL_CONFIG_1)
    cur &= ~0x11111000
    cur |= (value & 0x7) << 3
    self.setb(2, self.ACCEL_CONFIG_1, cur)

  accelFilterBWs = [265., 265., 136., 68.8, 34.4, 17., 8.3, 499.]
  @property
  def accelFilterBW(self):
    return self.accelFilterBWs[self.accelFilterConfig]
  @accelFilterBW.setter
  def accelFilterBW(self, value):
    (best, ind) = self.findClosest(self.accelFilterBWs, value)
    self.accelFilterConfig = ind

  @property
  def accelAvgConfig(self):
    return self.getb(2, self.ACCEL_CONFIG_2)
  @accelAvgConfig.setter
  def accelAvgConfig(self, value):
    return self.setb(2, self.ACCEL_CONFIG_2, value)

  @property
  def accelAvg(self):
    avg = 2**(self.accelAvgConfig + 2)
    if not self.accelUseFilter:
      avg = 1
    return avg
  @accelAvg.setter
  def accelAvg(self, value):
    self.accelAvgConfig = max(0, int(math.log2(value)) - 2) & 0x3

  def setAccelConfig(self):
    self.accelSampleRateDivider = self.accel_smplrt_div
    self.accelScale = self.accel_scale
    if self.raw_values:
      self.accel_factor = 1
    else:
      self.accel_factor = 2 * self.accelScale / (2**16)
      if not self.accel_g:
        self.accel_factor *= self.normalEarthAcc
    self.accelFilterBW = self.accel_filter_bw
    self.accelUseFilter = self.accel_use_filter
    self.accelAvg = self.accel_avg
    self.setupAccelCorrection()

  def setupAccelCorrection(self):
    if self.accel_corr:
      self.accel_offsets = self.accelOffsetsForScale
    else:
      self.accel_offsets = [0, 0, 0]

  def getAccelValueUnCorr(self, val):
    val *= self.accel_factor
    return val

  def getAccelValues(self, vals):
    return [self.getAccelValueUnCorr(vals[i] - self.accel_offsets[i])*self.accel_ax_factors[i] for i in range(3)]

  def readAccel(self):
    ax, ay, az = self.spi.get3x2(0x2D)
    ax = self.getAccelValueUnCorr(ax - self.accel_offsets[0])*self.accel_ax_factors[0]
    ay = self.getAccelValueUnCorr(ay - self.accel_offsets[1])*self.accel_ax_factors[1]
    az = self.getAccelValueUnCorr(az - self.accel_offsets[2])*self.accel_ax_factors[2]
    return (ax, ay, az)

  length_mag_data = 8

  def setupMagReadout(self):
    self.i2c0_setRegReadout(0xC, 0x11, self.length_mag_data)

  mag_scale = 4.9152 # Millitesla
  mag_factor = 2 * mag_scale / (2**16)
  def getMagValue(self, val):
    val *= self.mag_factor
    return val

  def getMagValues(self, vals):
    return tuple(self.getMagValue(val) for val in vals)

  def readMag(self):
    ax, ay, az = self.spi.get3x2(self.EXT_SENS_DATA_00, '<')
    return self.getMagValues((ax, ay, az))


  def read9D(self):
    return self.readAccel() + self.readGyro() + self.readMag()

  Temp_Sensitivity = 333.87
  RoomTemp_Offset = 0

  TEMP_OUT_H = 0x39
  def readTemp(self):
    v = self.get2(self.TEMP_OUT_H)
    return self.getTempValue(v)

  def getTempValue(self, val):
    return ((val - self.RoomTemp_Offset)/self.Temp_Sensitivity) + 21


  def readConfigured(self):
    return self.readAccel() + self.readGyro() + (self.readTemp(),)


  FIFO_EN_1     = 0x66
  FIFO_EN_2     = 0x67
  FIFO_RST      = 0x68
  FIFO_MODE     = 0x69
  FIFO_COUNT_H  = 0x70
  FIFO_COUNT_L  = 0x71
  FIFO_R_W      = 0x72
  DATA_RDY_STATUS=0x74
  FIFO_CFG      = 0x76
  INT_STATUS_2  = 0x1B
  INT_STATUS_3  = 0x1C

  @property
  def resetFIFO(self):
    return (self.get(self.FIFO_RST) & 0xF) != 0
  @resetFIFO.setter
  def resetFIFO(self, b):
    if b:
      self.set(self.FIFO_RST, 0xF)
    else:
      self.set(self.FIFO_RST, 0)

  @property
  def enableSlave0FIFO(self):
    return self.getBit(self.FIFO_EN_1, 0)
  @enableSlave0FIFO.setter
  def enableSlave0FIFO(self, b):
    self.setBit(self.FIFO_EN_1, 0, b)

  @property
  def enableAccelFIFO(self):
    return self.getBit(self.FIFO_EN_2, 4)
  @enableAccelFIFO.setter
  def enableAccelFIFO(self, b):
    self.setBit(self.FIFO_EN_2, 4, b)

  @property
  def enableGyroXFIFO(self):
    return self.getBit(self.FIFO_EN_2, 1)
  @enableGyroXFIFO.setter
  def enableGyroXFIFO(self, b):
    self.setBit(self.FIFO_EN_2, 1, b)
  @property
  def enableGyroYFIFO(self):
    return self.getBit(self.FIFO_EN_2, 2)
  @enableGyroYFIFO.setter
  def enableGyroYFIFO(self, b):
    self.setBit(self.FIFO_EN_2, 2, b)
  @property
  def enableGyroZFIFO(self):
    return self.getBit(self.FIFO_EN_2, 3)
  @enableGyroZFIFO.setter
  def enableGyroZFIFO(self, b):
    self.setBit(self.FIFO_EN_2, 3, b)

  @property
  def enableTempFIFO(self):
    return self.getBit(self.FIFO_EN_2, 0)
  @enableTempFIFO.setter
  def enableTempFIFO(self, b):
    self.setBit(self.FIFO_EN_2, 0, b)

  def setupFIFO(self):
    self.enableI2CCycle = False
    self.enableAccelCycle = False
    self.enableGyroCycle = False

    self.enableAccelFIFO = self.fifo_accel
    self.disableAccel = not self.fifo_accel

    self.enableGyroXFIFO = self.enableGyroYFIFO = self.enableGyroZFIFO = self.fifo_gyro
    self.disableGyro = not self.fifo_gyro

    self.enableTempFIFO = self.fifo_temp

    self.set(self.FIFO_MODE, 0) # 0 stream, 1 snapshot
    self.set(self.FIFO_CFG, 0) # interupts off

    self.enableFIFO = True
    self.resetFIFO = True

  @property
  def FIFOCount(self):
    return self.get2(self.FIFO_COUNT_H) & 0xFFF
  @property
  def FIFOOverflow(self):
    return self.get(self.INT_STATUS_2) & 0x1
  @property
  def FIFOWatermarkReached(self):
    return self.get(self.INT_STATUS_3) & 0x1

  @property
  def FIFOState(self):
    fcnt = self.FIFOCount
    wmfl = self.FIFOWatermarkReached
    ovfl = self.FIFOOverflow
    if fcnt > 0:
      # can't be
      ovfl = 0
    return (fcnt, wmfl, ovfl)

  def getFIFOByte(self):
    return self.get(self.FIFO_R_W)

  def getFIFOBytes(self, N):
    return self.spi.getbytes(self.FIFO_R_W, N)

  def getFIFOValues(self, vals):
    res = []
    offs = 0

    if self.fifo_accel:
      res += self.getAccelValues(bytes2words(vals[offs:offs+6]))
      offs += 6
    if self.fifo_gyro:
      res += self.getGyroValues(bytes2words(vals[offs:offs+6]))
      offs += 6
    if self.fifo_temp:
      res += [self.getTempValue(bytes2words(vals[offs:offs+2])[0])]
      offs += 2
    if self.fifo_mag:
      res += self.getMagValues(bytes2words(vals[offs:offs+self.length_mag_data]))
      offs += self.length_mag_data
    return res

  def getFIFOBlockSize(self):
    offs = 0
    if self.enableAccelFIFO:
      offs += 6
    if self.enableGyroXFIFO:
      offs += 6
    self.bank = 0
    return offs

  I2C_MST_ODR_CONFIG = 0x00
  I2C_MST_CTRL       = 0x01
  I2C_MST_DELAY_CTRL = 0x02
  I2C_SLV0_ADDR      = 0x03
  I2C_SLV0_REG       = 0x04
  I2C_SLV0_CTRL      = 0x05


  I2C_MST_STATUS     = 0x17
  EXT_SENS_DATA_00   = 0x3B

  @property
  def i2cMasterODRConfig(self):
    return self.getb(3, self.I2C_MST_ODR_CONFIG)
  @i2cMasterODRConfig.setter
  def i2cMasterODRConfig(self, val):
    self.setb(3, self.I2C_MST_ODR_CONFIG, val & 0xF)
  @property
  def i2cMasterODR(self):
    return 1125 / (2**self.i2cMasterODRConfig)
  @i2cMasterODR.setter
  def i2cMasterODR(self, val):
    divi = 1125 / val
    self.i2cMasterODRConfig = int(round(math.log2(divi)))

  @property
  def enableI2CMultiMaster(self):
    self.bank = 3
    b = self.getBit(self.I2C_MST_CTRL, 7)
    self.bank = 0
    return b
  @enableI2CMultiMaster.setter
  def enableI2CMultiMaster(self, val):
    self.bank = 3
    self.setBit(self.I2C_MST_CTRL, 7, val)
    self.bank = 0

  @property
  def enableI2CStopRestart(self):
    self.bank = 3
    v = self.getBit(self.I2C_MST_CTRL, 4)
    self.bank = 0
    return 'stop' if v else 'restart'
  @enableI2CStopRestart.setter
  def enableI2CStopRestart(self, val):
    if val == 'stop':
      val = 1
    elif val == 'restart':
      val = 0
    self.setBit(self.I2C_MST_CTRL, 4, val)
  @property
  def i2cMasterClockConfig(self):
    v = self.getb(3, self.I2C_MST_CTRL)
    return v & 0xF


  def setI2CConfig(self):
    icm20649 = self
    icm20649.set(icm20649.I2C_MST_ODR_CONFIG, 0x1)
    icm20649.set(icm20649.I2C_MST_CTRL, 0x7)
    icm20649.set(icm20649.I2C_MST_DELAY_CTRL, 0)

  def i2c0_setRegReadout(self, addr, reg, N):
    #print("i2c0 get addr 0x%x reg 0x%x" % (addr, reg))
    self.setUserBank(3)
    icm20649 = self
    icm20649.set(icm20649.I2C_SLV0_ADDR, 0x80 | addr)
    icm20649.set(icm20649.I2C_SLV0_REG,  reg)
    # read, N B
    assert(N < 0x10)
    icm20649.set(icm20649.I2C_SLV0_CTRL, 0x80 | (N & 0xF))
    self.setUserBank(0)

  def i2c0_cancelRegReadout(self):
    #print("i2c0 get addr 0x%x reg 0x%x" % (addr, reg))
    self.setUserBank(3)
    self.set(self.I2C_SLV0_CTRL, 0)
    self.setUserBank(0)

  def i2c0_getReg(self, addr, reg):
    #print("i2c0 get addr 0x%x reg 0x%x" % (addr, reg))
    self.setUserBank(3)
    icm20649 = self
    icm20649.set(icm20649.I2C_SLV0_ADDR, 0x80 | addr)
    icm20649.set(icm20649.I2C_SLV0_REG,  reg)
    # read, 1 B
    icm20649.set(icm20649.I2C_SLV0_CTRL, 0x80 | 1)

    rval = None
    for i in range(10):
      dr = self.get(icm20649.I2C_SLV0_CTRL)
      print("i2c0 getreg: 0x%x" %(dr,))
      if dr & 0x80 == 0:
        break
      time.sleep(1e-4)

    self.setUserBank(0)

    print("i2c0 getreg: 0x%x" %(dr,))
    if dr & 0x80 == 0:
      rval = icm20649.get(icm20649.EXT_SENS_DATA_00)
      print("i2c0 get addr 0x%x reg 0x%x = 0x%x" % (addr, reg, rval))
    return rval


  I2C_SLV4_ADDR      = 0x13
  I2C_SLV4_REG       = 0x14
  I2C_SLV4_CTRL      = 0x15
  I2C_SLV4_DO        = 0x16
  I2C_SLV4_DI        = 0x17

  def i2c4_setReg(self, addr, reg, val):
    #print("i2c0 get addr 0x%x reg 0x%x" % (addr, reg))
    self.setUserBank(3)
    icm20649 = self
    icm20649.set(icm20649.I2C_SLV4_ADDR, addr)
    icm20649.set(icm20649.I2C_SLV4_REG,  reg)
    icm20649.set(icm20649.I2C_SLV4_DO,  val)
    self.set(self.I2C_SLV4_CTRL, 0x80)

    for i in range(10):
      sr = self.get(self.I2C_SLV4_CTRL)
      if sr & 0x80 == 0:
        break
      else:
        print("4+")
        pass
      time.sleep(1e-4)

    print("i2c4 set addr 0x%x reg 0x%x = 0x%x" % (addr, reg, val))
    self.setUserBank(0)

  def i2c4_getReg(self, addr, reg):
    #print("i2c0 get addr 0x%x reg 0x%x" % (addr, reg))
    self.setUserBank(3)
    icm20649 = self
    icm20649.set(icm20649.I2C_SLV4_ADDR, 0x80 | addr)
    icm20649.set(icm20649.I2C_SLV4_REG,  reg)
    val = icm20649.i2c4_readReg()
    if val is not None:
      #print("i2c4 get addr 0x%x reg 0x%x = 0x%x" % (addr, reg, val))
      pass
    return val

  def i2c4_readReg(self):
    self.set(self.I2C_SLV4_CTRL, 0x80)

    for i in range(10):
      sr = self.get(self.I2C_SLV4_CTRL)
      if sr & 0x80 == 0:
        break
      else:
        #print("4.")
        pass
      time.sleep(1e-4)


    rval = None
    if sr & 0x80 == 0:
      rval = self.get(self.I2C_SLV4_DI)

    self.setUserBank(0)

    return rval

  def whoami(self):
    assert(self.bank == 0)
    return self.get(0)

  @property
  def connected(self):
    return self.whoami() == valWhoami

  INT_STATUS_1 = 0x1A

  @property
  def dataReady(self):
    return self.getBit(self.INT_STATUS_1, 0)

  def wait4data(self, timeout=1000):
    t0 = time.time()
    while 1e3 * (time.time() - t0) < timeout:
      d = self.dataReady
      if d:
        #print("data ready", d)
        break
      time.sleep(5e-4)

  GYRO_SMPLRT_DIV = 0x00
  GYRO_CONFIG_1  = 0x01
  GYRO_CONFIG_2  = 0x02

  ODR_ALIGN_EN      = 0x9
  ACCEL_SMPLRT_DIV_1= 0x10
  ACCEL_SMPLRT_DIV_2= 0x11
  ACCEL_CONFIG_1 = 0x14
  ACCEL_CONFIG_2 = 0x15

  USER_CTRL      = 0x03

  @property
  def enableODRAlign(self):
    self.bank = 2
    b = self.getBit(self.ODR_ALIGN_EN, 0)
    self.bank = 0
    return b
  @enableODRAlign.setter
  def enableODRAlign(self, b):
    self.bank = 2
    self.setBit(self.ODR_ALIGN_EN, 0, b)
    self.bank = 0


  XA_OFFS_H = 0x14
  YA_OFFS_H = 0x17
  ZA_OFFS_H = 0x1A

  XG_OFFS_USRH = 0x03
  YG_OFFS_USRH = 0x05
  ZG_OFFS_USRH = 0x07

  accel_corr = 1
  gyro_corr = 1
  gyro_correction_file = 'calibg.json'

  accel_offsets = [0, 0, 0]
  gyro_offsets = [0, 0, 0]

  gyrox_offset = -11.8
  gyroy_offset = -19.42
  gyroz_offset =  13.58

  @property
  def accelOffsets(self):
    self.bank = 1
    res = [self.get2(self.XA_OFFS_H), self.get2(self.YA_OFFS_H), self.get2(self.ZA_OFFS_H)]
    self.bank = 0
    return res

  @property
  def accelOffsetsForScale(self):
    res = self.accelOffsets
    s = self.accelScale
    res = [v / (s/2) for v in res]
    return res

  @property
  def gyroOffsets(self):
    self.bank = 2
    res = [self.get2(self.XG_OFFS_USRH), self.get2(self.YG_OFFS_USRH), self.get2(self.ZG_OFFS_USRH)]
    self.bank = 0
    return res

  def loadGyroCorrection(self):
    fname = self.gyro_correction_file
    print('Load gyro correction values from "%s"' % (fname,))
    with open(fname) as inf:
      import json
      data = json.loads(inf.read())
      self.gyrox_offset = data['gyro_offset'][0]
      self.gyroy_offset = data['gyro_offset'][1]
      self.gyroz_offset = data['gyro_offset'][2]
      print('Loaded gyro correction values: %s' % (data,))

  def setupGyroCorrection(self):
    if self.gyro_corr == 0:
      self.gyro_offsets = [0, 0, 0]
    elif self.gyro_corr == 1:
      # from device
      self.gyro_offsets = self.gyroOffsets
    elif self.gyro_corr == 2:
      # from config file
      self.loadGyroCorrection()
      self.gyro_offsets = [self.gyrox_offset, self.gyroy_offset, self.gyroz_offset]
    s = 2**self.gyroScaleConfig
    self.gyro_offsets = [v/s for v in self.gyro_offsets]

  def setup(self):
    icm20649 = self

    icm20649.enableSleep = False
    icm20649.enableLowPower = False

    icm20649.enableI2CCycle = self.i2c_cycle
    icm20649.enableAccelCycle = self.accel_cycle
    icm20649.enableGyroCycle = self.gyro_cycle

    print(self.temp_disable, type(self.temp_disable))
    if self.accel_disable:
      self.disableAccel = True
    if self.gyro_disable:
      self.disableGyro = True
    if self.temp_disable:
      self.disableTemperature = True

    self.enableODRAlign = self.enable_odr_align

    icm20649.setGyroConfig()
    icm20649.setAccelConfig()

    self.tempFilterBW = self.temp_filter_bw

    #    icm20649.setUserBank(3)
    #    icm20649.setI2CConfig()

    self.useTimeCorr = True

    self.timeScale = 1
    if self.useTimeCorr:
      self.timeScale += self.timebaseCorrection * 1e-2

    self.bank = 0
    icm20649.enableLowPower = self.low_power_mode

    t0 = time.time()
    for i in range(100):
      d = icm20649.get(0x1A)
      if d:
        t1 = time.time()
        print("data 0x%x ready after %g ms" %(d, 1e3 * (t1 - t0)))
        break
      time.sleep(1e-4)
    if not d:
      print("data not ready at setup", d)


def parseargs():
  import argparse
  import configparser

  conf_parser = argparse.ArgumentParser(add_help=False)
  conf_parser.add_argument("-c", "--conf_file", default=["/etc/icm20649.conf", "icm20649.conf"],
                           help="Specify config file", metavar="FILE")
  conf_parser.add_argument("-C", "--conf_section", default="icmsensor",
                           help="Specify config file section", metavar="NAME")
  args, remaining_argv = conf_parser.parse_known_args()

  def_dev = 0
  try:
    def_dev_env = os.getenv('ICM20649_SPI_DEV')
    if def_dev_env:
      def_dev = int(def_dev_env)
  except BaseException as e:
    print(e)
    pass

  defaults = {
    'debug': 0,

    'bus': 0,
    'dev': def_dev,
    'speed': int(2e6),       # Hz

    'maxreads': -1,
    'maxreadt': -1,
    'raw_values': 0,
    'ilff': 1,
    'output': 'lstofile',
    'nclients': 1,

    'outfile': 'out.csv',
    'downsample_outfile': '',
    'downsample_rate': 100,
    'precision': 8,
    'format': 'eng',

    'fifo_accel': 1,
    'fifo_gyro': 1,
    'fifo_mag': 0,
    'fifo_temp': 0,

    'i2c_cycle': 0,

    'accel_corr': 1,
    'accelx_factor': 1,
    'accely_factor': 1,
    'accelz_factor': 1,
    'accel_cycle': 0,
    'accel_disable': 0,
    'accel_freq': 1125,      # Hz
    'accel_scale': 4,        # g
    'accel_use_filter': 1,
    'accel_filter_bw': 500,  # Hz
    'accel_avg': 1,
    'accel_g': 1,

    'gyro_corr': 1,
    'gyro_correction_file': 'calibg.json',
    'gyro_cycle': 0,
    'gyro_disable': 0,
    'gyro_freq': 1125,       # Hz
    'gyro_scale': 2000,      # deg/s (dps)
    'gyro_use_filter': 0,
    'gyro_filter_bw': 375,   # Hz
    'gyro_avg': 1,

    'temp_disable': 1,
    'temp_filter_bw': 8000,

    'temp_ext_enable': '0,1',
    'temp_ext_b': 3500,

    'tempmanrfreq': 10,
    'magmanrfreq': 100

  }
  if args.conf_file:
    config = configparser.ConfigParser(defaults=defaults)
    print("Reading configuration file from %s" % (args.conf_file,))
    cread = config.read(args.conf_file)
    print("Read configuration files: %s" % (cread,))
    if len(cread):
      defaults = dict(config.items(args.conf_section))
      # List values need conversion str -> list
      if len(defaults['temp_ext_enable']):
        defaults['temp_ext_enable'] = [int(v) for v in defaults['temp_ext_enable'].split(',')]
      else:
        del defaults['temp_ext_enable']
      print('Loading configuration defaults from %s/[%s]' % (cread, args.conf_section))
      print('defaults', defaults)

  parser = argparse.ArgumentParser(description='Read ICM20649 sensor',
                                   parents=[conf_parser],
                                   formatter_class=argparse.RawDescriptionHelpFormatter)

  parser.add_argument('--debug', type=int, help='Debug')
  parser.add_argument('--bus', dest='bus', type=int, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, help='SPI speed')

  parser.add_argument('--maxreads', type=int, help='number of reads')
  parser.add_argument('--maxreadt', type=float, help='time limit of reading')
  parser.add_argument('--raw-values', dest='raw_values', type=int, help='output raw 16 bit values')
  parser.add_argument('--rel-times', dest='rel_times', type=int, help='output relative time stamps')
  parser.add_argument('--ilff', type=int, help='Output using ILFF')
  parser.add_argument('--output', type=str, help='Output mode: lstofile, socket, file')
  parser.add_argument('--nclients', type=int, help='Number of socket listeners')

  parser.add_argument('--fifo-accel',  dest='fifo_accel',  type=int, help='Accel. to FIFO')
  parser.add_argument('--fifo-gyro',   dest='fifo_gyro',   type=int, help='Gyro. to FIFO')
  parser.add_argument('--fifo-mag',    dest='fifo_mag',    type=int, help='Magn. to FIFO')
  parser.add_argument('--fifo-temp',   dest='fifo_temp',   type=int, help='Temp. to FIFO')

  parser.add_argument('--gyro-corr',   type=int, help='Use Gyro correction')
  parser.add_argument('--gyro-correction-file', type=str, help='Gyro correction values JSON')
  parser.add_argument('--gyro-cycle',   type=int, help='Duty-cycle Gyro')
  parser.add_argument('--gyro-disable',  dest='gyro_disable',   type=int, help='Disable Gyro')
  parser.add_argument('--gyro-freq',   dest='gyro_freq',   type=float, help='Gyro frequency')
  parser.add_argument('--gyro-scale',  dest='gyro_scale',  type=float, help='Gyro scale: 250, 500, 1000, 2000')
  parser.add_argument('--gyro-filter', dest='gyro_use_filter', type=int,   help='Gyro filter: 0, 1')
  parser.add_argument('--gyro-filter-bw', dest='gyro_filter_bw', type=float,   help='Gyro filter BW: 8.9, 17.8, ..., 229.8 Hz')
  parser.add_argument('--gyro-avg',    dest='gyro_avg',    type=float, help='Gyro avg.: 1, 2, 4, ..., 128')

  parser.add_argument('--accel-corr',   type=int, help='Use Accel correction')
  parser.add_argument('--accelx-factor',   type=int, help='Factor for Accel X axis')
  parser.add_argument('--accely-factor',   type=int, help='Factor for Accel Y axis')
  parser.add_argument('--accelz-factor',   type=int, help='Factor for Accel Z axis')
  parser.add_argument('--accel-cycle',   type=int, help='Duty-cycle Accel')
  parser.add_argument('--accel-disable',   dest='accel_disable',   type=int, help='Disable Accel')
  parser.add_argument('--accel-freq',  dest='accel_freq',  type=float, help='Accel frequency')
  parser.add_argument('--accel-scale', dest='accel_scale', type=float, help='Accel scale: 4, 8, 16, 30')
  parser.add_argument('--accel-filter',dest='accel_use_filter',type=int,   help='Accel filter: 0, 1')
  parser.add_argument('--accel-filter-bw',dest='accel_filter_bw',type=float,   help='Accel filter BW: 8.3, 17.0, ..., 499 Hz')
  parser.add_argument('--accel-avg',   dest='accel_avg',   type=float, help='Accel scale: 2, 4, 8, 16')
  parser.add_argument('--accel-g',   dest='accel_g',   type=int, help='Accel unit: 0: m/s, 1: g')

  parser.add_argument('--temp-disable',   dest='temp_disable',   type=int, help='Disable Temperature')
  parser.add_argument('--temp-filter-bw',dest='temp_filter_bw',type=float,   help='Temp filter BW: 8.3, 17.0, ..., 499 Hz')

  parser.add_argument('--temp-ext-enable', type=int, nargs='*', help='Temp manual retrieval freq (Hz)')
  parser.add_argument('--temp-ext-freq', dest='tempmanrfreq',type=float,   help='Temp manual retrieval freq (Hz)')
  parser.add_argument('--temp-ext-b', type=float,  help='Temp ext sensor param B')

  parser.add_argument('--mag-freq', dest='magmanrfreq',type=float,   help='Magnet. manual retrieval freq (Hz)')

  parser.add_argument('--outfile', '-o', type=str, help='Output CSV file')
  parser.add_argument('--downsample-outfile', dest='downsample_outfile', type=str, help='Downsample CSV output file')
  parser.add_argument('--downsample-rate', dest='downsample_rate', type=int, help='Downsample CSV output with rate')
  parser.add_argument('--precision', '-p', type=int, help='CSV precision')
  parser.add_argument('--format', type=str, help='CSV value format (fixed or adaptive)')

  parser.set_defaults(**defaults)

  args = parser.parse_args(remaining_argv)

  print('parseargs', vars(args))
  return args


def getICM20649(init=True):
  import json
  args = parseargs()
  print('Configuration:', json.dumps(vars(args), indent=1))

  spi = SPIConn(args.bus, args.dev, args.speed)
  print('SPI', spi)

  icm20649 = ICM20649(spi, vars(args))

  if init:
    icm20649.reset()
    icm20649.setup()

  return (icm20649, args)
