#! /usr/bin/env python3

from spitool import SPIConn
import time
import argparse
import math
import struct
import os
import sys
import log
import subprocess

from icm20649 import getICM20649
import table
import ilff
import linesocket

regWhoami = 0
valWhoami = 0xEA

regReset = 0x06
valReset = 0x41

if __name__ == "__main__":
  log.name = 'icm20649_fifo'

  if False:
    pid = 0
    affinity = os.sched_getaffinity(pid)
    print("Process is eligible to run on:", affinity)
    affinity = os.sched_setaffinity(pid, {0})
    affinity = os.sched_getaffinity(pid)
    print("Process is now eligible to run on:", affinity)

  icm20649, args = getICM20649()

  fifodata = b''
  nfifodata = 0

  rawWords = []
  nreads = 0
  nsamples = 0
  blksz = icm20649.getFIFOBlockSize()
  process = True

  smplrt = icm20649.sampleRate
  smpldt = 1 / smplrt

  icm20649.setupFIFO()

  icm20649.bank = 0
  print(icm20649)

  header = ('Time (s)',)
  if not icm20649.disableAccel:
    header += ('Acc_x (g)', 'Acc_y (g)', 'Acc_z (g)')
  if not icm20649.disableGyro:
    header += ('Gyr_x (dps)', 'Gyr_y (dps)', 'Gyr_z (dps)')
  if icm20649.fifo_temp or icm20649.tempmanrfreq > 0:
    header += ('Temp. (°C)',)

  log.msg('Open output mode "%s", file name "%s"' % (args.output, args.outfile))
  if args.output == 'lstofile':
    outf = ls = linesocket.LineSocket()
    lstfcmd = ['python3', './icm20649_spi/lstofile.py',
               '--infile=%s' % (ls.server_address,), '--outfile=%s' % (args.outfile,),
               '--debug=%d' % (args.debug,), '--ilff=%d' % (args.ilff,)]
    print('Launching lstofile helper process: %s' % (' '.join(lstfcmd),))
    p = subprocess.Popen(lstfcmd, shell=False)
    ls.openSocket(1)
  elif args.output == 'socket':
    outf = ls = linesocket.LineSocket()
    ls.openSocket(args.nclients)
  elif args.output == 'ilff':
    outf = ilff.ILFFFile(args.outfile, 'w')
  else:
    outf = open(args.outfile, 'w')

  print('Use output:', outf)

  precs = args.precision
  if not args.rel_times:
    precs = [16] + [precs]*(len(header)-1)
  outt = table.Table(outf, header, precs)
  outt.writeHeader()

  iloop = 0
  i0 = ilast = 0
  t0 = tlast = tManTempLast = tManMagLast = time.time()

  icm20649.resetFIFO = True
  icm20649.enableFIFO = False
  icm20649.resetFIFO = False
  icm20649.enableFIFO = True

  fcnt, wmfl, ovfl = icm20649.FIFOState
  avgfcnt = fcnt
  stdfcnt = 0
  log.msg("FIFO start: count %d, wmfl %d, ovfl %d" % (fcnt, wmfl, ovfl))

  maxfcnt = 0
  nresets = 0
  nwmfl = 0
  avgAlpha = 0.9

  curTempVals = [0]

  if icm20649.tempmanrfreq > 0:
    tempManInter = 1 / icm20649.tempmanrfreq
    curTempVals = [icm20649.readTemp(),]

  while True:

    timeval = time.time()
    telapse = timeval - t0
    telapseR = timeval - tlast
    tlast = timeval
    timevalOut = telapse if args.rel_times else timeval

    fcnts = []
    while True:
      fcnt, wmfl, ovfl = icm20649.FIFOState
      maxfcnt = max(maxfcnt, fcnt)
      avgfcnt = avgfcnt * avgAlpha + fcnt * (1 - avgAlpha)
      stdfcnt = (fcnt - avgfcnt)
      fcnts += [fcnt]

      v = icm20649.getFIFOBytes(fcnt)
      fifodata += struct.pack('B'*fcnt, *v)
      nfifodata += fcnt

#      sys.stdout.write("FIFO: %s %d %d\r" % (fcnts, wmfl, ovfl))
#      sys.stdout.flush()

      if ovfl:
        icm20649.resetFIFO = 1
        icm20649.resetFIFO = 0
        fcnt_, wmfl_, ovfl_ = icm20649.FIFOState
        log.error("FIFO overflow: it %d (now: %d/%d/%d), FIFO %d, buf %d" %
                  (iloop, fcnt_, wmfl_, ovfl_, fcnt, len(fifodata),))
        nresets += 1
        break
      if wmfl:
        nwmfl += 1
        log.error("FIFO watermark reached: it %d, FIFO %d, buf %d" % (iloop, fcnt, len(fifodata)))
      if fcnt < 200:
        break

    if iloop % 400 == 0:
      sys.stdout.write("%.5f %.5f: FIFO: %d (avg %.1f %+.1f, max %d), ovfl %d/%d wm %d/%d: %d B, %.2f KB/s, CSV %d B %.3f MB/s, Buf %d B\n" %
                       (telapse, telapseR, fcnt, avgfcnt, stdfcnt, maxfcnt, ovfl, nresets, wmfl, nwmfl,
                        nfifodata, 1e-3 * nfifodata / telapse, outt.nchar, 1e-6 * outt.nchar / telapse, len(fifodata) ))
      sys.stdout.flush()

    sideFetch = False
    if fcnt < 1000 and len(fifodata) < 100000 and not ovfl and icm20649.tempmanrfreq > 0 and timeval - tManTempLast > tempManInter:
      sideFetch = True
      curTempVals = [icm20649.readTemp(),]
      tManTempLast = timeval

    if process and not ovfl and not wmfl:

      nblk_poss = int(len(fifodata)/blksz)
      nblk = min(40, nblk_poss)

      for i in range(nblk):
        block = fifodata[0:blksz]
        block = icm20649.getFIFOValues(block)
        blktm = timevalOut - (nblk - i - 1) * smpldt
        #print(timeval, (nblk - i - 1), blktm)
        block = [blktm,] + block
        if icm20649.tempmanrfreq > 0:
          block += curTempVals
        try:
          outt.writeValues(block)
        except BlockingIOError:
          log.msg('socket currently full: FIFO %d, Buffer %d' % (fcnt, len(fifodata)))
          time.sleep(1e-3)
          break
        fifodata = fifodata[blksz:]

      nsamples += nblk*blksz
      nreads += nblk

    if args.maxreads >= 0 and nreads > args.maxreads:
      print("\nReached max read samples: %g" % (args.maxreads,))
      break
    if args.maxreadt >= 0 and telapse > args.maxreadt:
      print("\nReached max read time: %g" % (args.maxreadt,))
      break
    iloop += 1
    if fcnt < 1000 and not sideFetch:
      time.sleep(1e-4)

  icm20649.enableFIFO = False

  print("read %d samples from %d blocks from queue, wrote %d B CSV data" % (nsamples, nreads, outt.nchar))

# Local Variables:
# coding: utf-8
# End:
