#! /usr/bin/env python3

import time
import argparse
import math
import struct
import os
import sys
import asyncio
import functools
import concurrent

from spitool import SPIConn
from icm20948 import getICM20948
import table
import ilff
import log
from timer import WTimer, Timer

class Status:
  fcnt = wmfl = ovfl = 0
  maxfcnt = avgfcnt = stdfcnt = nfifodata = 0
  nsamples = nreads = nresets = nwmfl = 0
  curMagVals = [0, 0, 0]
  curTempVals = [0]

class Context:
  done = False
  que = None
  args = None
  sensor = None

status = Status()
context = Context()

executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
def run_in_executor(f):
    @functools.wraps(f)
    def inner(*args, **kwargs):
        loop = asyncio.get_running_loop()
        return loop.run_in_executor(executor, lambda: f(*args, **kwargs))

    return inner


def setupSensor():
  icm20948, args = getICM20948()

  icm20948.setupFIFO()
  if icm20948.magmanrfreq > 0:
    print('setup mag side channel reading')
    icm20948.enableI2CCycle = True

  icm20948.bank = 0

  return icm20948, args


@run_in_executor
def writeCSValues(outt, batch):
  for block in batch:
    outt.writeValues(block)

avgAlpha = 0.8
@run_in_executor
def getFIFOBatch(icm20948, timer):
  status.fcnt, status.wmfl, status.ovfl = icm20948.FIFOState
  timer.click()

  status.maxfcnt = max(status.maxfcnt, status.fcnt)
  status.avgfcnt = status.avgfcnt * avgAlpha + status.fcnt * (1 - avgAlpha)
  status.stdfcnt = (status.fcnt - status.avgfcnt)

  if status.fcnt == 0:
    return b''

  v = icm20948.getFIFOBytes(status.fcnt)
  fifodata = struct.pack('B'*status.fcnt, *v)
  status.nfifodata += status.fcnt
  return fifodata

async def readFIFOSPI(context):
  print('readFIFOSPI start')
  que = context.que
  icm20948 = context.sensor
  args = context.args

  iloop = 0

  if icm20948.tempmanrfreq > 0:
    tempManInter = 1 / icm20948.tempmanrfreq
    status.curTempVals = [icm20948.readTemp(),]

  if icm20948.magmanrfreq > 0:
    magManInter = 1 / icm20948.magmanrfreq
    status.curMagVals = icm20948.readMag()

  timer = WTimer()

  icm20948.enableFIFO = False
  icm20948.enableFIFO = True
  icm20948.resetFIFO = True
  status.fcnt, status.wmfl, status.ovfl = icm20948.FIFOState
  icm20948.resetFIFO = False

  while True:
    iloop += 1

    fifodata = await getFIFOBatch(icm20948, timer)

    if len(fifodata) == 0:
      if context.done:
        break
      else:
        await asyncio.sleep(1e-3)
        continue

    await que.put((timer.elaps if args.rel_times else timer.time, fifodata))

    if status.ovfl:
      icm20948.resetFIFO = 1
      icm20948.resetFIFO = 0
      fcnt_, wmfl_, ovfl_ = icm20948.FIFOState
      log.error("FIFO overflow: it %d (now: %d/%d/%d), FIFO %d, buf %d" %
                (iloop, fcnt_, wmfl_, ovfl_, status.fcnt, len(fifodata),))
      status.nresets += 1
    if status.wmfl:
      status.nwmfl += 1
      log.error("FIFO watermark reached: it %d, FIFO %d, buf %d" % (iloop, status.fcnt, len(fifodata)))

    sidefetch = False
    if status.fcnt < 1000 and icm20948.tempmanrfreq > 0 and timer.elapsed('tempfetch', tempManInter):
      status.curTempVals = [icm20948.readTemp(),]
      sidefetch = True
    if status.fcnt < 1000 and icm20948.magmanrfreq > 0 and timer.elapsed('magfetch', magManInter):
      status.curMagVals = icm20948.readMag()
      sidefetch = True

    if status.fcnt < 200 and not sidefetch:
      await asyncio.sleep(2e-3)

  print('readFIFOSPI stop')


async def processValues(context):
  print('processValues start')

  icm20948 = context.sensor
  args = context.args
  que = context.que

  header = ('Time (s)',)
  if not icm20948.disableAccel:
    header += ('Acc_x (g)', 'Acc_y (g)', 'Acc_z (g)')
  if not icm20948.disableGyro:
    header += ('Gyr_x (dps)', 'Gyr_y (dps)', 'Gyr_z (dps)')
  if icm20948.fifo_temp or icm20948.tempmanrfreq > 0:
    header += ('Temp. (°C)',)
  if icm20948.fifo_mag or icm20948.magmanrfreq > 0:
    header += ('Mag_x (µT)', 'Mag_y (µT)', 'Mag_z (µT)')

  log.msg('Open output mode "%s", file name "%s"' % (args.output, args.outfile))
  if args.ilff:
    outf = ilff.ILFFFile(args.outfile, 'w')
  else:
    outf = open(args.outfile, 'w')

  print('Use output:', outf)
  sys.stdout.flush()

  precs = args.precision
  if not args.rel_times:
    precs = [16] + [precs]*(len(header)-1)
  outt = table.Table(outf, header, precs)
  outt.writeHeader()

  blksz = icm20948.getFIFOBlockSize()

  smplrt = icm20948.sampleRate
  smpldt = 1 / smplrt

  timer = WTimer()

  batch = []

  while not context.done or not que.empty():
    timer.click()

    (tstamp, fifodata) = await que.get()

    nblk = int(len(fifodata)/blksz)

    for i in range(nblk):
      block = fifodata[0:blksz]
      fifodata = fifodata[blksz:]
      block = icm20948.getFIFOValues(block)
      blktm = tstamp - (nblk - i - 1) * smpldt
      block = [blktm,] + block
      if icm20948.tempmanrfreq > 0:
        block += status.curTempVals
      if icm20948.magmanrfreq > 0:
        block += status.curMagVals
      batch += [block]

    if len(batch) > 200:
      await writeCSValues(outt, batch)
      batch = []

    status.nsamples += nblk*blksz
    status.nreads += nblk

    if timer.elapsed('info', 1):
      print("%s: FIFO: %d (avg %.1f %+.1f, max %d), ovfl %d/%d wm %d/%d: %d B, %.2f KB/s, CSV %d B %.3f MB/s, Queue %d" %
                       (timer, status.fcnt, status.avgfcnt, status.stdfcnt, status.maxfcnt,
                        status.ovfl, status.nresets, status.wmfl, status.nwmfl,
                        status.nfifodata, 1e-3 * status.nfifodata / timer.elaps,
                        outt.nchar, 1e-6 * outt.nchar / timer.elaps, que.qsize() ))
      sys.stdout.flush()

    if args.maxreads >= 0 and status.nreads > args.maxreads and not context.done:
      print("\nReached max read samples: %g" % (args.maxreads,))
      icm20948.enableFIFO = False
      icm20948.resetFIFO = True
      context.done = True
    if args.maxreadt >= 0 and timer.elaps > args.maxreadt and not context.done:
      print("\nReached max read time: %g" % (args.maxreadt,))
      icm20948.enableFIFO = False
      icm20948.resetFIFO = True
      context.done = True

  await writeCSValues(outt, batch)
  batch = []

  print('processValues stop')


async def main():
  log.name = 'icm20948_fifo_as'

  if False:
    pid = 0
    affinity = os.sched_getaffinity(pid)
    print("Process is eligible to run on:", affinity)
    affinity = os.sched_setaffinity(pid, {0})
    affinity = os.sched_getaffinity(pid)
    print("Process is now eligible to run on:", affinity)

  icm20948, args = setupSensor()
  print(icm20948)

  context.que = asyncio.Queue()
  context.args = args
  context.sensor = icm20948

  # Run the tasks
  with Timer(text="\nTotal elapsed time: {:.1f}"):
    await asyncio.gather(
      asyncio.create_task(readFIFOSPI(context)),
      asyncio.create_task(processValues(context)),
    )

  assert(context.done == True)

  print("read %d samples from %d blocks from queue" % (status.nsamples, status.nreads))

if __name__ == "__main__":
  asyncio.run(main())

# Local Variables:
# coding: utf-8
# End:
