#! /usr/bin/env python3

# python module
import time
import argparse
import math
import struct
import os
import sys
import subprocess
import socket
import fcntl
import termios
import threading
import signal

# fionec modules
import ilff

# package module
from icm20649_spi.spitool import SPIConn
from icm20649_spi.icm20649 import getICM20649
from icm20649_spi import table
from icm20649_spi import log
from icm20649_spi.timer import WTimer
from icm20649_spi import decimate
from icm20649_spi import monitor
from icm20649_spi import senseHatTemperature as st

class Status:
  fcnt = wmfl = ovfl = 0
  maxfcnt = avgfcnt = stdfcnt = nfifodata = 0
  nsamples = nreads = nresets = nwmfl = 0
  curTempVals = []
  nsockdata = 0
  subpid = 0
  subproc = None

class Context:
  done = False
  sock = None
  args = None
  sensor = None
  params = None
  decim = None
  outt = None

status = Status()

def setupSensor():
  icm20649, args = getICM20649()

  icm20649.setupFIFO()

  icm20649.bank = 0

  return icm20649, args


def initOutput(context):
  icm20649 = context.sensor
  args = context.args

  if icm20649.temp_ext_enable:
    channels = [st.avail_channels[i] for i in icm20649.temp_ext_enable]
    opts = {'constant_B': args.temp_ext_b}
    st.init(channels, **opts)
    status.curTempVals = st.readall()
  else:
    st.init([])

  header = ('Time (s)',)
  if not icm20649.disableAccel:
    header += ('Acceleration_X (g)', 'Acceleration_Y (g)', 'Acceleration_Z (g)')
  if not icm20649.disableGyro:
    header += ('Rotation_X (°/s)', 'Rotation_Y (°/s)', 'Rotation_Z (°/s)')
  if icm20649.fifo_temp:
    header += 'Temperature_0 (°C)'
  if icm20649.temp_ext_enable:
    header += tuple('Temperature_%d (°C)' % (i+1,) for i in icm20649.temp_ext_enable)

  log.msg('Open output mode "%s", file name "%s"' % (args.output, args.outfile))
  if args.ilff:
    outf = ilff.ILFFFile(args.outfile, 'w')
    print('using ILFF version %s' % (ilff.VERSION))
  else:
    outf = open(args.outfile, 'w')

  precs = args.precision
  if not args.rel_times:
    precs = [16] + [precs]*(len(header)-1)
  outt = table.Table(outf, header, precs)
  outt.writeHeader()

  tdefFile = 'tabledef.json'
  with open(tdefFile, 'w') as f:
    f.write(repr(outt))

  if args.downsample_outfile == '':
    (a, b) = os.path.splitext(args.outfile)
    of100 = a + '_%dHz' % (args.downsample_rate) + b
  else:
    of100 = args.downsample_outfile

  smplrt = icm20649.sampleRate
  smpldt = 1 / smplrt

  dsFactor = int(round(smplrt / args.downsample_rate))

  xc_home = '.'
  try:
    xc_home_env = os.getenv('XC_HOME')
    if xc_home_env:
      xc_home = os.path.join(xc_home_env, 'etc')
  except:
    pass

  outf100 = open(of100, 'w')
  outt100 = outt.getV3Table(outf100)
  outt100.writeHeader()

  xc_home = '.'
  xc_home_etc = '.'
  try:
    xc_home_env = os.getenv('XC_WORKDIR')
    if xc_home_env:
      xc_home = xc_home_env
      xc_home_etc = os.path.join(xc_home_env, 'etc')
  except:
    pass

  with open(os.path.join(xc_home, 'um-header.csv'), 'w') as f:
    f.write(outt.getHeader())

  monitor.actionsFilename = os.path.join(xc_home_etc, 'actions.json')
  monitor.bordersFilename = os.path.join(xc_home_etc, 'limits.json')
  monitor.eventsFilename = os.path.join(xc_home, 'events.csv')
  monitor.readHeaders(outt.headers)
  monitor.readBorders()
  monitor.startEventLog()
  monitor.cancelActions()

  outLatest = open(os.path.join(xc_home, 'um-latest.csv'), 'w')

  def writeSlowValues(items):
    ln = outt100.fmtValues(items)
    outLatest.seek(0)
    outLatest.truncate()
    outLatest.write(ln)
    outLatest.flush()
    outt100.writeLine(ln)

  decim = decimate.Decimator(dsFactor, len(header),
                             writeSlowValues,
                             lambda items, mode: monitor.processValues(items, mode))

  context.decim = decim
  context.outt = outt

def processValues(context):
  sock = context.sock
  decim = context.decim
  icm20649 = context.sensor
  outt = context.outt

  blksz = context.params['blksz']

  smplrt = context.params['smplrt']
  smpldt = 1 / (smplrt * icm20649.timeScale)

  iall = 0
  tbase = 0

  timer = WTimer()
  if not context.args.rel_times:
    tbase = timer.start

  log.msg('icm20649_fifo_mt queue read loop starting')

  sockdata = b''
  tempManInter = 1e-2

  while not context.done:
    timer.click()

    while True:
      try:
        chunk = sock.recv(4096)
        # log.msg('Got %d B chunk' % (len(chunk),))
        if len(chunk) == 0:
          context.done = True
          break
        sockdata += chunk
        if len(sockdata) > 1e3:
          break
      except BlockingIOError as ex:
        # log.msg('socket queue is empty')
        if len(sockdata):
          break
        else:
          time.sleep(1e-4)
      except KeyboardInterrupt as ex:
        print('Crtl-C')
        break
      except BaseException as ex:
        log.msg('Error reading socket: %s' % (ex,))
        raise(ex)
        break

    # log.msg('Got %d B of data from socket' % (len(sockdata),))
    nblk = int(len(sockdata)/blksz)
    status.nsockdata += len(sockdata)

    for i in range(nblk):
      block = sockdata[0:blksz]
      sockdata = sockdata[blksz:]
      block = icm20649.getFIFOValues(block)
      blktm = tbase + iall * smpldt
      block = [blktm,] + block
      if icm20649.tempmanrfreq > 0:
        block += status.curTempVals

      oln = outt.fmtValues(block)
      outt.writeLine(oln)

      decim.processValues(block)
      iall += 1

    if timer.elapsed('info', 1) or (timer.last > 40 and nblk > 2):
      nread = fcntl.ioctl(sock, termios.FIONREAD, b'\x00' * 4, 1)
      nread = struct.unpack('i', nread)[0]
      print("%s%s: %d B, %.2f KB/s, CSV %d B %.3f MB/s, pending %d B" %
            (">" if timer.last > 40 else " ", timer, status.nsockdata, 1e-3 * status.nsockdata / timer.elaps,
             outt.nchar, 1e-6 * outt.nchar / timer.elaps, nread))
      sys.stdout.flush()

    status.nsamples += nblk*blksz
    status.nreads += nblk

  log.msg('icm20649_fifo_mt queue read loop finished')
  monitor.stopActions()


def tempReadThreadFun(context, tempmanrfreq=1):
  tempManInter = 1 / tempmanrfreq
  nc = len(st.channels)
  if nc == 0:
    return
  tempManInter = tempManInter / nc
  print('tempReadThreadFun start:', nc, 'channels', tempmanrfreq, 'Hz', tempManInter, 's')
  timer = WTimer()
  ci = 0
  while not context.done:
    timer.click()
    if status.fcnt < 1000 and tempmanrfreq > 0 and timer.elapsed('tempfetch', tempManInter):
      status.curTempVals[ci] = st.read(ci)
      #print('Temp', ci, st.channels, st.ntc_param_B, st.res_R2, status.curTempVals[ci])
      ci += 1
      ci %= nc
    else:
      time.sleep(1e-3)
  print('tempReadThreadFun exit')

def startProcess(cargs):
  print('starting FIFO reader process')

  args = ['sudo', 'icm20649-fiforeader'] + cargs

  print('start process %s' % (args,))
  proc = subprocess.Popen(args)
  print('started process %s,%s' % (args,proc.pid))
  status.subpid = proc.pid
  status.subproc = proc

def ctrlCHandler(sig, frame):
  print('Signal %d caught, passing on to helper process %d' % (sig,status.subpid))
  if status.subpid:
    proc = subprocess.run(['sudo', 'killall', 'icm20649-fiforeader'])
    if proc.returncode != 0:
      print('Failed to kill helper')

def run():
  log.name = 'icm20649_fifo_mp'

  icm20649, args = setupSensor()
  print(icm20649)

  context = Context()
  context.args = args
  context.sensor = icm20649

  log.msg('icm20649_fifo_mt starting')
  log.msg('icm20649_fifo_mt starting helper thread')

  server_address = '/run/usense/linesocket-%d.sock' % (os.getpid(),)

  params = {}
  params['blksz'] = icm20649.getFIFOBlockSize()
  params['smplrt'] = icm20649.sampleRate
  print('params', params)

  context.params = params

  signal.signal(signal.SIGINT, ctrlCHandler)
  signal.signal(signal.SIGTERM, ctrlCHandler)

  initOutput(context)

  tempThread = threading.Thread(target=tempReadThreadFun,args=(context,10))
  print('start', tempThread)
  tempThread.start()

  startProcess(['--bus', '%d' % args.bus,
                '--dev', '%d' % args.dev,
                '--speed', '%d' % args.speed,
                '--outfile', server_address])

  sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
  sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, int(1e8))

  t0 = time.time()
  maxWaitTime = 5
  while True:
    log.msg('connecting to {}'.format(server_address))
    try:
      sock.connect(server_address)
      break
    except socket.error as msg:
      failmsg = msg
    if time.time() - t0 > maxWaitTime:
      log.msg('Failed to connect to server in %g s' % (maxWaitTime,))
      log.msg(failmsg)
      sys.exit(1)
    time.sleep(0.5)

  log.msg('connected to {}'.format(server_address))
  sys.stdout.flush()

  sock.setblocking(False)
  context.sock = sock

  log.msg('icm20649_fifo_mp starting main loop')
  try:
    processValues(context)
  except BaseException as ex:
    print(f'Error {ex} occured in main processing loop, exiting program')
    context.done = True
    raise(ex)

  tempThread.join()
  print("read %d samples from %d blocks from queue" % (status.nsamples, status.nreads))

if __name__ == "__main__":
  run()

# Local Variables:
# coding: utf-8
# End:
