#! /usr/bin/env python3

# python module
import time
import argparse
import math
import struct
import os
import sys
import threading
import queue

# fionec modules
import ilff

# package module
from icm20649_spi.spitool import SPIConn
from icm20649_spi.icm20649 import getICM20649
from icm20649_spi import table
from icm20649_spi import log
from icm20649_spi.timer import WTimer
from icm20649_spi import decimate
from icm20649_spi import monitor
from icm20649_spi import queuereader
from icm20649_spi import socketreader
from icm20649_spi import senseHatTemperature as st

useDecimatorSocket = True

class Status:
  fcnt = wmfl = ovfl = 0
  maxfcnt = avgfcnt = stdfcnt = nfifodata = 0
  nsamples = nreads = nresets = nwmfl = 0
  curTempVals = []

class Context:
  done = False
  que = None
  cmdque = None
  args = None
  sensor = None

status = Status()
context = Context()

def setupSensor():
  icm20649, args = getICM20649()

  icm20649.setupFIFO()

  icm20649.bank = 0

  return icm20649, args


def readFIFOSPI(context):
  print('readFIFOSPI start')

  que = context.que
  icm20649 = context.sensor

  while True:
    try:
      cmd = context.cmdque.get()
      if cmd == 'start':
        break
      else:
        print('error: strange start command: %s' % (cmd,))
    except queue.Empty:
      time.sleep(1e-3)
      pass

  print('read FIFO: start')
  assert(icm20649.bank == 0)

  iloop = 0
  avgAlpha = 0.8

  timer = WTimer()

  icm20649.resetFIFO = True
  status.fcnt, status.wmfl, status.ovfl = icm20649.FIFOState
  icm20649.resetFIFO = False
  assert(icm20649.bank == 0)

  fifodata = b''
  blksz = icm20649.getFIFOBlockSize()
  assert(icm20649.bank == 0)

  print('readFIFOSPI start II')

  maxReadBlkSize = 1024

  def getQueue(fetch=0):
    fcnt, wmfl, ovfl = icm20649.FIFOState
    timer.click()
    rcnt = int(min(fcnt,maxReadBlkSize)/blksz)*blksz
    if rcnt > 0:
      v = icm20649.getFIFOBytes(rcnt)
      fifodata = struct.pack('B'*rcnt, *v)
      status.nfifodata += rcnt
    else:
      fifodata = b''

    status.maxfcnt = max(status.maxfcnt, fcnt)
    tstmp = timer.elaps if context.args.rel_times else timer.time

    if ovfl:
      icm20649.resetFIFO = 1
      icm20649.resetFIFO = 0
      log.error("FIFO overflow: it %d/%d, FIFO %d, buf %d" % (iloop, fetch, fcnt, len(fifodata)))
      status.nresets += 1
      sys.stdout.flush()
    elif wmfl:
      status.nwmfl += 1
#      log.error("FIFO watermark reached: it %d/%d, FIFO %d, buf %d" % (iloop, fetch, fcnt, len(fifodata)))

    status.fcnt, status.wmfl, status.ovfl = (fcnt, wmfl, ovfl)
    return (fifodata, tstmp)

  while True:
    iloop += 1
    assert(icm20649.bank == 0)

    while True:
      (fifodata, tmstmp) = getQueue()
      if len(fifodata) > 0:
        que.put((tmstmp, fifodata), block=False)
      if len(fifodata) < 250:
        break

    if status.fcnt == 0:
      if context.done:
        break
      else:
        if not status.ovfl and not status.wmfl:
          time.sleep(1e-4)

#    if timer.elapsed('break', 150) and False:
#      print('break')
#      time.sleep(0.1)

    if status.fcnt < 200:
      time.sleep(2e-3)

    if que.qsize() > 10000:
      log.error('queue got too large (%d), stopping' % (que.qsize()))
      break

  print('readFIFOSPI stop')

def processValues(context):
  icm20649 = context.sensor
  args = context.args
  que = context.que

  channels = [st.avail_channels[i] for i in icm20649.temp_ext_enable]
  opts = {'constant_B': args.temp_ext_b}
  st.init(channels, **opts)
  status.curTempVals = st.readall()

  header = ('Time (s)',)
  if not icm20649.disableAccel:
    unit = 'g' if icm20649.accel_g else 'm/s²'
    header += ('Acceleration_X (%s)' % (unit,), 'Acceleration_Y (%s)' % (unit,), 'Acceleration_Z (%s)' % (unit,))
  if not icm20649.disableGyro:
    header += ('Rotation_X (°/s)', 'Rotation_Y (°/s)', 'Rotation_Z (°/s)')
  if icm20649.fifo_temp or icm20649.tempmanrfreq > 0:
    header += tuple('Temperature_%d (°C)' % (i+1,) for i in icm20649.temp_ext_enable)

  log.msg('Open output mode "%s", file name "%s"' % (args.output, args.outfile))
  if args.ilff:
    outf = ilff.ILFFFile(args.outfile, 'w')
    print('using ILFF version %s' % (ilff.VERSION))
  else:
    outf = open(args.outfile, 'w')

  precs = args.precision
  if not args.rel_times:
    precs = [16] + [precs]*(len(header)-1)
  outt = table.Table(outf, header, precs)

  tdefFile = 'tabledef.json'
  with open(tdefFile, 'w') as f:
    f.write(repr(outt))

  outf.close()

  if args.downsample_outfile == '':
    (a, b) = os.path.splitext(args.outfile)
    of100 = a + '_%dHz' % (args.downsample_rate) + b
  else:
    of100 = args.downsample_outfile

  smplrt = icm20649.sampleRate
  smpldt = 1 / (smplrt * icm20649.timeScale)

  print('Timescale (rate): %g * %g = %g' % (icm20649.sampleRate, icm20649.timeScale, icm20649.sampleRate * icm20649.timeScale))
  print('Timescale   (dt): %g / %g = %g' % (1 / icm20649.sampleRate, icm20649.timeScale, 1 / (icm20649.sampleRate * icm20649.timeScale)))

  dsFactor = int(round(smplrt / args.downsample_rate))

  xc_home = '.'
  xc_home_etc = '.'
  try:
    xc_home_env = os.getenv('XC_WORKDIR')
    if xc_home_env:
      xc_home = xc_home_env
      xc_home_etc = os.path.join(xc_home_env, 'etc')
  except:
    pass

  if useDecimatorSocket:
    subargs = ['--outfile', args.outfile]
    subargs += ['--outfile-slow', of100]
    subargs += ['--output', 'ilff' if args.ilff else args.output]
    subargs += ['--downsample-factor', '%d' % (dsFactor,)]
    subargs += ['--actions-file', os.path.join(xc_home_etc, 'actions.json')]
    subargs += ['--limits-file', os.path.join(xc_home_etc, 'limits.json')]
    subargs += ['--event-file', os.path.join(xc_home, 'events.csv')]
    subargs += ['--header-file', os.path.join(xc_home, 'um-header.csv')]
    subargs += ['--latest-file', os.path.join(xc_home, 'um-latest.csv')]
    subargs += ['--table-def-file', tdefFile]
    socketReader = socketreader.SocketReader()
    socketReader.startProcess(subargs)

  else:
    outf100 = open(of100, 'w')
    outt100 = outt.getV3Table(outf100)
    outt100.writeHeader()

    monitor.actionsFilename = 'actions.json'
    monitor.bordersFilename = 'limits.json'
    monitor.readHeaders(header)
    monitor.readBorders()
    monitor.cancelActions()

    decim = decimate.Decimator(dsFactor, len(header),
                               lambda items: outt100.writeValues(items),
                               lambda items: monitor.processValues(items))

    qreaderDecimateAndActions = queuereader.QueueReaderMT(lambda x: decim.processValues(x))

  blksz = icm20649.getFIFOBlockSize()

  timer = WTimer()
  fifodata = b''
  log.msg('icm20649_fifo_mt queue read loop started and ready')

  context.cmdque.put('start')

  tstamp0 = None
  nblkout = 0

  if icm20649.tempmanrfreq > 0:
    tempManInter = 1 / icm20649.tempmanrfreq

  while True:
    timer.click()

    try:
      (tstamp, fifodata) = que.get(False)
      if tstamp0 is None:
        tstamp0 = tstamp
      # time.sleep(1e-3)
    except queue.Empty:
      if context.done:
        print('Queue is empty and stop set')
        break
      else:
        time.sleep(2e-3)

    nblk = int(len(fifodata)/blksz)
    #print('Got %d B = %d blocks' % (len(fifodata), nblk))

    for i in range(nblk):
      block = fifodata[0:blksz]
      fifodata = fifodata[blksz:]
      block = icm20649.getFIFOValues(block)
      blktm = tstamp0 + (status.nreads + i) * smpldt
      block = [blktm,] + block
      if icm20649.tempmanrfreq > 0:
        block += status.curTempVals

      if useDecimatorSocket:
        oln = outt.fmtValues(block)
        socketReader.putLine(oln)
      else:
        qreaderDecimateAndActions.que.put(block)

    status.nsamples += nblk*blksz
    status.nreads += nblk

    if icm20649.tempmanrfreq > 0 and timer.elapsed('tempfetch', tempManInter):
      status.curTempVals = st.readall()

    if timer.elapsed('info', 1):
      print("%s%s: FIFO: %d (max %d), ov %d wm %d: %d B, %.2f KB/s, CSV %d B %.3f MB/s, Queue %d" %
                       (">" if timer.last > 40 else " ", timer, status.fcnt, status.maxfcnt,
                        status.nresets, status.nwmfl,
                        status.nfifodata, 1e-3 * status.nfifodata / timer.elaps,
                        outt.nchar, 1e-6 * outt.nchar / timer.elaps, que.qsize() ))
      sys.stdout.flush()
      if status.fcnt == 0:
        if timer.elapsed('infoempty', 10):
          print('FIFO is currently empty')

    if args.maxreads >= 0 and status.nreads > args.maxreads and not context.done:
      print("\nReached max read samples: %g" % (args.maxreads,))
#      icm20649.enableFIFO = False
      icm20649.resetFIFO = True
      context.done = True
    if args.maxreadt >= 0 and timer.elaps > args.maxreadt and not context.done:
      print("\nReached max read time: %g" % (args.maxreadt,))
#      icm20649.enableFIFO = False
      icm20649.resetFIFO = True
      context.done = True

  if not useDecimatorSocket:
    qreaderDecimateAndActions.done = context.done
  log.msg('icm20649_fifo_mt queue read loop finished')

def run():
  log.name = 'icm20649_fifo'

  if False:
    pid = 0
    affinity = os.sched_getaffinity(pid)
    print("Process is eligible to run on:", affinity)
    affinity = os.sched_setaffinity(pid, {0})
    affinity = os.sched_getaffinity(pid)
    print("Process is now eligible to run on:", affinity)

  icm20649, args = setupSensor()
  print(icm20649)

  context.que = queue.SimpleQueue()
  context.cmdque = queue.SimpleQueue()
  context.args = args
  context.sensor = icm20649

  log.msg('icm20649_fifo_mt starting')
  log.msg('icm20649_fifo_mt starting helper thread')

  rFIFOThread = threading.Thread(target=readFIFOSPI, args=(context,))
  rFIFOThread.start()

  log.msg('icm20649_fifo_mt starting main loop')
  processValues(context)

  assert(context.done == True)

  rFIFOThread.join()

  print("read %d samples from %d blocks from queue" % (status.nsamples, status.nreads))

if __name__ == "__main__":
  run()

# Local Variables:
# coding: utf-8
# End:
