#! /usr/bin/python3

import spidev
import time
import struct

import argparse

def readRegInt8(spi, addr):
  req = [addr | 0x80, 0]
  antwort = spi.xfer2(req)
#  print('RECV Byte %02x' % (addr,), ' '.join([('%02x' % (i,)) for i in antwort[1:]]))
  return struct.unpack('b', struct.pack('B', antwort[1]))[0]

def readRegUInt8(spi, addr):
  req = [addr | 0x80, 0]
  antwort = spi.xfer2(req)
#  print('RECV Byte %02x' % (addr,), ' '.join([('%02x' % (i,)) for i in antwort[1:]]))
  return antwort[1]

def writeReg8(spi, addr, data):
  req = [addr, data]
  antwort = spi.xfer2(req)
#  print('SEND Byte %02x' % (addr,), ' '.join([('%02x' % (i,)) for i in [data]]))
  return antwort[1]


def readRegInt16(spi, addr):
  req = [addr | 0x80, 0, 0]
  antwort = spi.xfer2(req)
#  print('RECV Word %02x' % (addr,), ' '.join([('%02x' % (i,)) for i in antwort[1:]]))
  return struct.unpack('h', struct.pack('BB', antwort[2], antwort[1]))[0]

def read3RegInt16(spi, addr, end='>'):
  req = [addr | 0x80, 0, 0, 0, 0, 0, 0]
  antwort = spi.xfer2(req)
#  print('RECV Word %02x' % (addr,), ' '.join([('%02x' % (i,)) for i in antwort[1:]]))
  return struct.unpack(end+'h'*3, struct.pack('B'*6, *antwort[1:]))

def readRegInt24(spi, addr):
  req = [addr | 0x80, 0, 0, 0]
  antwort = spi.xfer2(req)
  #print('RECV 3B %02x' % (addr,), ' '.join([('%02x' % (i,)) for i in antwort[1:]]))
  return struct.unpack('i', struct.pack('BBBB', antwort[1], antwort[2], antwort[3], 0))[0]

def readRegInt32(spi, addr):
  req = [addr | 0x80, 0, 0, 0, 0]
  antwort = spi.xfer2(req)
  #print('RECV DWORD %02x' % (addr,), ' '.join([('%02x' % (i,)) for i in antwort[1:]]))
  return struct.unpack('i', struct.pack('BBBB', antwort[1], antwort[2], antwort[3], antwort[4]))[0]


class SPIConn:

  spi = None

  def __init__(self, bus, dev, busSpeed=5000):
    self.spi = spidev.SpiDev()
    self.bus = bus
    self.dev = dev
    self.spi.open(bus,dev)
    self.spi.max_speed_hz = busSpeed
    self.spi.mode = 0b11

  def __str__(self):
    return "SPI(%d, %d), speed %d" % (self.bus, self.dev, self.spi.max_speed_hz)

  def readRegUInt8(self, addr):
    return readRegUInt8(self.spi, addr)

  def readRegInt8(self, addr):
    return readRegInt8(self.spi, addr)

  def writeReg8(self, addr, data):
    return writeReg8(self.spi, addr, data)

  def writeRegUInt8(self, addr, data):
    return writeReg8(self.spi, addr, data)

  def set(self, addr, data):
    return self.writeReg8(addr, data)

  def setr(self, addr, data):
    self.writeReg8(addr, data)
    return self.get(addr)

  def get(self, addr):
    return self.readRegUInt8(addr)

  def gets(self, addr):
    return self.readRegInt8(addr)

  def readRegInt16(self, addr):
    return readRegInt16(self.spi, addr)

  def get2(self, addr):
    return self.readRegInt16(addr)

  def get3x2(self, addr, end='>'):
    return read3RegInt16(self.spi, addr, end)

  def readRegInt24(self, addr):
    return readRegInt24(self.spi, addr)

  def get3(self, addr):
    return self.readRegInt24(addr)

  def readRegInt32(self, addr):
    return readRegInt32(self.spi, addr)

  def get4(self, addr):
    return self.readRegInt32(addr)

  def getbytes(self, addr, ln):
    req = [addr | 0x80] + [0]*ln
    antwort = self.spi.xfer2(req)
    return antwort[1:ln+1]


def parseargs():
  parser = argparse.ArgumentParser(description='Erzeuge Umspul-Konfiguration.')

  parser.add_argument('--bus', dest='bus', type=int, default=0, help='SPI bus')
  parser.add_argument('--dev', dest='dev', type=int, default=0, help='SPI device')
  parser.add_argument('--speed', dest='speed', type=int, default=int(1e6), help='SPI speed')

  parser.add_argument('--word', dest='word', action='store_const',
                      const=True, default=False, help='Word transaction')
  parser.add_argument('--dword', dest='dword', action='store_const',
                      const=True, default=False, help='dWord transaction')

  parser.add_argument('--set', dest='set', action='store_const',
                      const=True, default=False, help='Set register')
  parser.add_argument('--get', dest='get', action='store_const',
                      const=True, default=False, help='Get register')

  parser.add_argument('--mask', dest='mask', type=str, default='', help='Value mask')

  parser.add_argument('reg', metavar='reg', type=str, nargs=1,
                    help='Register address and value')

  parser.add_argument('val', metavar='val', type=str, nargs='?',
                    help='Register value')

  args = parser.parse_args()
  return args

def main():
  args = parseargs()

  spiconn = SPIConn(args.bus, args.dev, args.speed)

  reg = int(args.reg[0], 16)

  if args.set:
    if args.val[0:2] == '0x':
      val = int(args.val, 16)
    elif args.val[0:2] == '0b':
      val = int(args.val, 2)
    else:
      val = int(args.val)
#    print('Set %s: %s' % (reg, val))
    reply = spiconn.setr(reg, val)
    print('Set 0x%x: %s = %s' % (reg, val, reply))

  if args.get:
    if args.word:
      val = spiconn.readRegInt16(reg)
    elif args.dword:
      val = spiconn.readRegInt32(reg)
    else:
      val = spiconn.readRegUInt8(reg)
    if args.mask:
      val &= int(args.mask, 16)
    print('Get 0x%x: %d (0x%x, %s)' % (reg, val, val, bin(val)))


if __name__ == "__main__":
  main()
